import os
import abc
import json
import time
from datetime import datetime

import appdirs  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional, Generic, TypeVar

StateType = TypeVar('StateType')


class CacheStateError(BioLibError):
    pass


class CacheState(abc.ABC, Generic[StateType]):
    _cache_dir: str = appdirs.user_cache_dir(appname='pybiolib', appauthor='biolib')

    @property
    @abc.abstractmethod
    def _state_path(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_default_state(self) -> StateType:
        raise NotImplementedError

    @property
    def _state_lock_path(self) -> str:
        return f'{self._state_path}.lock'

    def __init__(self):
        self._state: Optional[StateType] = None

    def __enter__(self) -> StateType:
        logger_no_user_data.debug(f'CacheState: Entering state path: {self._state_path}...')
        self._acquire_state_lock()
        if os.path.exists(self._state_path):
            with open(self._state_path, mode='r') as file:
                self._state = json.loads(file.read())
        else:
            self._state = self._get_default_state()
            with open(self._state_path, mode='w') as file:
                file.write(json.dumps(self._state))

        # Check for type checking
        if self._state is None:
            raise CacheStateError('Internal state is not defined')

        return self._state

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        with open(self._state_path, mode='w') as file:
            file.write(json.dumps(self._state))

        self._release_state_lock()
        logger_no_user_data.debug(f'CacheState: Exited state path: {self._state_path}')

    def _acquire_state_lock(self) -> None:
        timeout_seconds = 5.0
        seconds_to_sleep = 0.5

        while os.path.exists(self._state_lock_path):
            time.sleep(seconds_to_sleep)
            timeout_seconds -= seconds_to_sleep
            if timeout_seconds < 0:
                raise CacheStateError('Cache state timed out waiting for lock file')

        os.makedirs(self._cache_dir, exist_ok=True)
        lock_file = open(self._state_lock_path, mode='x')
        lock_file.close()

    def _release_state_lock(self) -> None:
        if os.path.exists(self._state_lock_path):
            os.remove(self._state_lock_path)
        else:
            raise CacheStateError('Cache state was not locked.')

    @staticmethod
    def get_timestamp_now() -> str:
        return datetime.now().isoformat()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Worker module."""

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List

from pyee import AsyncIOEventEmitter
from biolib.pyppeteer.pyppeteer.execution_context import ExecutionContext
from biolib.pyppeteer.pyppeteer.jshandle import JSHandle
from biolib.pyppeteer.pyppeteer.models import JSFunctionArg

if TYPE_CHECKING:
    from pyppeteer.connection import CDPSession

logger = logging.getLogger(__name__)


class Worker(AsyncIOEventEmitter):
    """The Worker class represents a WebWorker.

    The events `workercreated` and `workerdestroyed` are emitted on the page
    object to signal the worker lifecycle.

    .. code::

        page.on('workercreated', lambda worker: print('Worker created:', worker.url))
    """

    def __init__(
        self,
        client: 'CDPSession',
        url: str,
        consoleAPICalled: Callable[[str, List[JSHandle], Any], None],
        exceptionThrown: Callable[[Dict], None],
    ) -> None:
        super().__init__()
        self._client = client
        self._url = url
        self.loop = client.loop
        self._executionContextPromise = self.loop.create_future()

        def jsHandleFactory(remoteObject: Dict) -> JSHandle:
            return None  # type: ignore

        def onExecutionContentCreated(event: Dict) -> None:
            nonlocal jsHandleFactory

            # noinspection PyRedeclaration
            def jsHandleFactory(remoteObject: Dict) -> JSHandle:
                return JSHandle(executionContext, client, remoteObject)

            executionContext = ExecutionContext(client, event['context'], None)
            self._executionContextCallback(executionContext)

        self._client.once('Runtime.executionContextCreated', onExecutionContentCreated)
        try:
            # This might fail if the target is closed before we receive all
            # execution contexts.
            self._client.send('Runtime.enable', {})
        except Exception as e:
            logger.error(f'An exception occurred: {e}')

        def onConsoleAPICalled(event: Dict[str, Any]) -> None:
            args: List[JSHandle] = []
            for arg in event.get('args', []):
                args.append(jsHandleFactory(arg))
            consoleAPICalled(event['type'], args, event['stackTrace'])

        self._client.on('Runtime.consoleAPICalled', onConsoleAPICalled)
        self._client.on(
            'Runtime.exceptionThrown', lambda exception: exceptionThrown(exception['exceptionDetails']),
        )

    def _executionContextCallback(self, value: ExecutionContext) -> None:
        self._executionContextPromise.set_result(value)

    @property
    def url(self) -> str:
        """Return URL."""
        return self._url

    @property
    async def executionContext(self) -> ExecutionContext:
        """Return ExecutionContext."""
        return await self._executionContextPromise

    async def evaluate(self, pageFunction: str, *args: JSFunctionArg) -> Any:
        """Evaluate ``pageFunction`` with ``args``.

        Shortcut for ``(await worker.executionContext).evaluate(pageFunction, *args)``.
        """
        return await (await self._executionContextPromise).evaluate(pageFunction, *args)

    async def evaluateHandle(self, pageFunction: str, *args: JSFunctionArg) -> JSHandle:
        """Evaluate ``pageFunction`` with ``args`` and return :class:`~pyppeteer.execution_context.JSHandle`.

        Shortcut for ``(await worker.executionContext).evaluateHandle(pageFunction, *args)``.
        """
        return await (await self._executionContextPromise).evaluateHandle(pageFunction, *args)



import os
import re
from pathlib import Path
import yaml

from biolib.lfs import get_iterable_zip_stream, get_files_and_size_of_directory
from biolib.typing_utils import Optional, Set
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.utils import get_absolute_container_image_uri

REGEX_MARKDOWN_INLINE_IMAGE = re.compile(r'!\[(?P<alt>.*)\]\((?P<src>.*)\)')


def push_application(
        app_uri: str,
        app_path: str,
        app_version_to_copy_images_from: Optional[str],
        is_dev_version: Optional[bool],
):
    api_client = BiolibApiClient.get()
    if not api_client.is_signed_in:
        # TODO: Create an exception class for expected errors like this that does not print stacktrace
        raise Exception(
            'You must be authenticated to push an application.\n'
            'Please set the environment variable "BIOLIB_TOKEN=[your_api_token]"\n'
            'You can get an API token at: https://biolib.com/settings/api-tokens/'
        ) from None

    # prepare zip file
    config_yml_path = Path(f'{app_path}/.biolib/config.yml')
    if not config_yml_path.is_file():
        raise BioLibError('The file .biolib/config.yml was not found in the application directory')

    zip_filters: Set[str] = set()
    zip_filters.add('.biolib/config.yml')

    try:
        with open(config_yml_path, mode='r') as config_yml_file:
            config = yaml.safe_load(config_yml_file.read())

        license_file_relative_path = config.get('license_file', 'LICENSE')
        if Path(app_path).joinpath(license_file_relative_path).is_file():
            zip_filters.add(license_file_relative_path)

        description_file_relative_path = config.get('description_file', 'README.md')
        description_file_full_path = Path(app_path).joinpath(description_file_relative_path)
        if not description_file_full_path.is_file():
            raise BioLibError(f'Could not find {description_file_relative_path}')

        zip_filters.add(description_file_relative_path)
        with open(description_file_full_path, mode='r') as description_file:
            description_file_content = description_file.read()

        for _, img_src_path in re.findall(REGEX_MARKDOWN_INLINE_IMAGE, description_file_content):
            zip_filters.add(img_src_path)

        for _, module in config['modules'].items():
            if module.get('source_files'):
                zip_filters.add('*')
                break

    except Exception as error:
        raise BioLibError('Failed to parse the .biolib/config.yml file') from error

    files_in_app_dir, _ = get_files_and_size_of_directory(directory=app_path)
    files_to_zip: Set[str] = set()

    for file_path in files_in_app_dir:
        for pattern in zip_filters:
            if pattern == '*' or pattern == file_path:
                files_to_zip.add(file_path)
                break

    original_directory = os.getcwd()
    app_path_obj = Path(app_path)
    os.chdir(app_path_obj.parent)
    files_with_app_dir_prefixed = [f'{app_path_obj.stem}/{path}' for path in files_to_zip]

    # Workaround as backend currently expects directory objects for root level and .biolib directory
    files_with_app_dir_prefixed.append(f'{app_path_obj.stem}/')
    files_with_app_dir_prefixed.append(f'{app_path_obj.stem}/.biolib/')

    byte_iterator = get_iterable_zip_stream(files_with_app_dir_prefixed, chunk_size=50_000_000)
    source_files_zip_bytes = b''.join(byte_iterator)
    os.chdir(original_directory)

    # TODO: Raise exception if app_uri contains version

    if app_version_to_copy_images_from and app_version_to_copy_images_from != 'active':
        # Get app with `app_version_to_copy_images_from` in app_uri to get the app version public id.
        app_uri += f':{app_version_to_copy_images_from}'

    app_response = BiolibAppApi.get_by_uri(app_uri)
    app = app_response['app']
    # push new app version
    new_app_version_json = BiolibAppApi.push_app_version(
        app_id=app['public_id'],
        app_name=app['name'],
        author=app['account_handle'],
        set_as_active=False,
        zip_binary=source_files_zip_bytes,
        app_version_id_to_copy_images_from=app_response['app_version']['public_id'] if app_version_to_copy_images_from
        else None
    )

    #  Don't push docker images if copying from another app version
    docker_tags = new_app_version_json.get('docker_tags', {})
    if not app_version_to_copy_images_from and docker_tags:
        logger.info('Found docker images to push.')

        try:
            yaml_file = open(f'{app_path}/.biolib/config.yml', 'r', encoding='utf-8')

        except Exception as error:  # pylint: disable=broad-except
            raise BioLibError('Could not open the config file .biolib/config.yml') from error

        try:
            config_data = yaml.safe_load(yaml_file)

        except Exception as error:  # pylint: disable=broad-except
            raise BioLibError('Could not parse .biolib/config.yml. Please make sure it is valid YAML') from error

        # Auth to be sent to proxy
        # The tokens are sent as "{access_token},{job_id}". We leave job_id blank on push.
        tokens = f'{BiolibApiClient.get().access_token},'
        auth_config = {'username': 'biolib', 'password': tokens}

        docker_client = BiolibDockerClient.get_docker_client()

        for module_name, repo_and_tag in docker_tags.items():
            docker_image_definition = config_data['modules'][module_name]['image']
            repo, tag = repo_and_tag.split(':')

            if docker_image_definition.startswith('dockerhub://'):
                docker_image_name = docker_image_definition.replace('dockerhub://', 'docker.io/', 1)
                logger.info(
                    f'Pulling image {docker_image_name} defined on module {module_name} from Dockerhub.')
                dockerhub_repo, dockerhub_tag = docker_image_name.split(':')
                docker_client.images.pull(dockerhub_repo, tag=dockerhub_tag, platform='linux/amd64')

            elif docker_image_definition.startswith('local-docker://'):
                docker_image_name = docker_image_definition.replace('local-docker://', '', 1)

            try:
                logger.info(f'Trying to push image {docker_image_name} defined on module {module_name}.')
                image = docker_client.images.get(docker_image_name)
                image_uri = get_absolute_container_image_uri(
                    base_url=api_client.base_url,
                    relative_image_uri=repo
                )
                image.tag(image_uri, tag)
                for line in docker_client.images.push(image_uri, tag=tag, stream=True,
                                                      decode=True, auth_config=auth_config):
                    logger.info(line)

            except Exception as exception:  # pylint: disable=broad-except
                raise BioLibError(f'Failed to tag and push image {docker_image_name}.') from exception

            logger.info(f'Successfully pushed {docker_image_name}')

        logger.info('Successfully pushed all docker images')

    if not is_dev_version:
        # Set new app version as active
        BiolibAppApi.update_app_version(
            app_version_id=new_app_version_json['public_id'],
            data={
                'set_as_active': True
            }
        )



import asyncio
import copy
import logging
import math
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import BrowserError, ElementHandleError, NetworkError
from biolib.pyppeteer.pyppeteer.models import JSFunctionArg, MouseButton, Protocol

if TYPE_CHECKING:
    from pyppeteer.page import Page
    from pyppeteer.frame import FrameManager
    from pyppeteer.execution_context import ExecutionContext


logger = logging.getLogger(__name__)


def createJSHandle(context, remoteObject) -> Union['JSHandle', 'ElementHandle']:
    frame = context.frame
    if remoteObject.get('subtype') == 'node' and frame:
        frameManager = frame._frameManager
        return ElementHandle(context, context._client, remoteObject, frameManager.page, frameManager)
    return JSHandle(context, context._client, remoteObject,)


class JSHandle:
    """JSHandle class.

    JSHandle represents an in-page JavaScript object. JSHandle can be created
    with the :meth:`~pyppeteer.page.Page.evaluateHandle` method.
    """

    def __init__(self, context: 'ExecutionContext', client: 'CDPSession', remoteObject: Protocol.Runtime.RemoteObject):
        self._context = context
        self._client = client
        self._remoteObject = remoteObject
        self._disposed = False

    @property
    def executionContext(self):
        """Get execution context of this handle."""
        return self._context

    async def evaluate(self, pageFunction: str, *args: JSFunctionArg):
        return await self.executionContext.evaluate(pageFunction, self, *args)

    async def evaluateHandle(self, pageFunction: str, *args: JSFunctionArg):
        return await self.executionContext.evaluateHandle(pageFunction, self, *args)

    async def getProperty(self, propertyName: str) -> 'JSHandle':
        """Get property value of ``propertyName``."""
        objectHandle = await self._context.evaluateHandle(
            '''(object, propertyName) => {
                const result = {__proto__: null};
                result[propertyName] = object[propertyName];
                return result;
            }''',
            self,
            propertyName,
        )
        properties = await objectHandle.getProperties()
        result = properties[propertyName]
        await objectHandle.dispose()
        return result

    async def getProperties(self) -> Dict[str, 'JSHandle']:
        """Get all properties of this handle."""
        response = await self._client.send(
            'Runtime.getProperties', {'objectId': self._remoteObject.get('objectId', ''), 'ownProperties': True,}
        )
        result = {}
        for prop in response['result']:
            if not prop.get('enumerable'):
                continue
            result[prop['name']] = createJSHandle(self._context, prop['value'])
        return result

    async def jsonValue(self) -> Dict:
        """Get Jsonized value of this object."""
        objectId = self._remoteObject.get('objectId')
        if objectId:
            response = await self._client.send(
                'Runtime.callFunctionOn',
                {
                    'functionDeclaration': 'function() { return this; }',
                    'objectId': objectId,
                    'returnByValue': True,
                    'awaitPromise': True,
                },
            )
            return helpers.valueFromRemoteObject(response['result'])
        return helpers.valueFromRemoteObject(self._remoteObject)

    def asElement(self) -> 'ElementHandle':
        ...

    async def dispose(self) -> None:
        """Stop referencing the handle."""
        if self._disposed:
            return
        self._disposed = True
        await helpers.releaseObject(self._client, self._remoteObject)

    def toString(self) -> str:
        """Get string representation."""
        if self._remoteObject.get('objectId'):
            _type = self._remoteObject.get('subtype') or self._remoteObject.get('type')
            return f'JSHandle@{_type}'
        return 'JSHandle:{}'.format(helpers.valueFromRemoteObject(self._remoteObject))


class ElementHandle(JSHandle):
    def __init__(
        self,
        context: 'ExecutionContext',
        client: CDPSession,
        remoteObject: Protocol.Runtime.RemoteObject,
        page: 'Page',
        frameManager: 'FrameManager',
    ):
        super().__init__(context, client, remoteObject)
        self._page = page
        self._frameManager = frameManager
        self._disposed = False

        # Aliases for query methods:
        self.J = self.querySelector
        self.Jx = self.xpath
        self.Jeval = self.querySelectorEval
        self.JJ = self.querySelectorAll
        self.JJeval = self.querySelectorAllEval

    def asElement(self) -> Optional['ElementHandle']:
        return self

    async def contentFrame(self):
        nodeInfo = await self._client.send('DOM.describeNode', {'objectId': self._remoteObject.get('objectId')})
        frameId = nodeInfo.get('node', {}).get('frameId')
        if isinstance(frameId, str):
            return self._frameManager.frame(frameId)

    async def _scrollIntoViewIfNeeded(self):
        error = await self.evaluate(
            """
            async(element, pageJavascriptEnabled) => {
              if (!element.isConnected)
                return 'Node is detached from document';
              if (element.nodeType !== Node.ELEMENT_NODE)
                return 'Node is not of type HTMLElement';
              // force-scroll if page's javascript is disabled.
              if (!pageJavascriptEnabled) {
                element.scrollIntoView({block: 'center', inline: 'center', behavior: 'instant'});
                return false;
              }
              const visibleRatio = await new Promise(resolve => {
                const observer = new IntersectionObserver(entries => {
                  resolve(entries[0].intersectionRatio);
                  observer.disconnect();
                });
                observer.observe(element);
              });
              if (visibleRatio !== 1.0)
                element.scrollIntoView({block: 'center', inline: 'center', behavior: 'instant'});
              return false;
            }
            """,
            self._page._javascriptEnabled,
        )
        if error:
            raise BrowserError(error)

    async def _clickablePoint(self):
        # swallow any errors to get a better error message
        # before:  Protocol error (DOM.getContentQuads): Could not compute content quads.
        # after swallowing: Node is either invisible or not an HTMLElement
        async def silent_get_quads():
            try:
                return await self._client.send('DOM.getContentQuads', {'objectId': self._remoteObject['objectId']})
            except NetworkError:
                logger.exception('Failed to retrieve content quads')
                return None

        result, layoutMetrics = await asyncio.gather(silent_get_quads(), self._client.send('Page.getLayoutMetrics'),)
        if not result or not result.get('quads'):
            raise BrowserError('Node is either not visible or not an HTMLElement')
        clientWidth = layoutMetrics['layoutViewport']['clientWidth']
        clientHeight = layoutMetrics['layoutViewport']['clientHeight']
        quads = []
        for quad in result['quads']:
            quad = self._intersectQuadWithViewport(self._fromProtocolQuad(quad), clientWidth, clientHeight)
            if computeQuadArea(quad) > 1:
                quads.append(quad)
        if not quads:
            raise BrowserError('Node is either not visible or not an HTMLElement')
        # return middle of quad[0]
        quad = quads[0]
        x, y = 0, 0
        for point in quad:
            x += point['x']
            y += point['y']
        return {
            'x': x / 4,
            'y': y / 4,
        }

    async def _getBoxModel(self):
        try:
            return await self._client.send('DOM.getBoxModel', {'objectId': self._remoteObject['objectId']})
        except Exception as e:
            logger.error(f'An exception occurred: {e}')

    def _fromProtocolQuad(self, quad):
        return [
            {'x': quad[0], 'y': quad[1]},
            {'x': quad[2], 'y': quad[3]},
            {'x': quad[4], 'y': quad[5]},
            {'x': quad[6], 'y': quad[7]},
        ]

    def _intersectQuadWithViewport(self, quad: List[Dict[str, float]], width: float, height: float):
        return [{'x': min(max(point['x'], 0), width), 'y': min(max(point['y'], 0), height)} for point in quad]

    async def hover(self) -> None:
        """Move mouse over to center of this element.

        If needed, this method scrolls element into view. If this element is
        detached from DOM tree, the method raises an ``ElementHandleError``.
        """
        await self._scrollIntoViewIfNeeded()
        obj = await self._clickablePoint()
        x = obj.get('x', 0)
        y = obj.get('y', 0)
        await self._page.mouse.move(x, y)

    async def click(self, button: MouseButton = 'left', clickCount: int = 1, delay: float = 0) -> None:
        """Click the center of this element.

        If needed, this method scrolls element into view. If the element is
        detached from DOM, the method raises ``ElementHandleError``.

        ``options`` can contain the following fields:

        * ``button`` (str): ``left``, ``right``, of ``middle``, defaults to
          ``left``.
        * ``clickCount`` (int): Defaults to 1.
        * ``delay`` (int|float): Time to wait between ``mousedown`` and
          ``mouseup`` in milliseconds. Defaults to 0.
        """
        await self._scrollIntoViewIfNeeded()
        point = await self._clickablePoint()
        x = point.get('x', 0)
        y = point.get('y', 0)
        await self._page.mouse.click(x, y, button=button, clickCount=clickCount, delay=delay)

    async def select(self, *values: str) -> List[str]:
        for val in values:
            if not isinstance(val, str):
                raise ValueError(f'value "{val}" needs to be of type str, but found value of type {type(val)}')
        return await self.evaluate(
            """(element, values) => {
              if (element.nodeName.toLowerCase() !== 'select')
                throw new Error('Element is not a <select> element.');

              const options = Array.from(element.options);
              element.value = undefined;
              for (const option of options) {
                option.selected = values.includes(option.value);
                if (option.selected && !element.multiple)
                  break;
              }
              element.dispatchEvent(new Event('input', { bubbles: true }));
              element.dispatchEvent(new Event('change', { bubbles: true }));
              return options.filter(option => option.selected).map(option => option.value);
            }
            """,
            values,
        )

    async def uploadFile(self, *filePaths: Union[Path, str]) -> dict:
        """Upload files."""
        # TODO port this
        files = [os.path.abspath(p) for p in filePaths]
        objectId = self._remoteObject.get('objectId')
        return await self._client.send('DOM.setFileInputFiles', {'objectId': objectId, 'files': files})

    async def tap(self) -> None:
        """Tap the center of this element.

        If needed, this method scrolls element into view. If the element is
        detached from DOM, the method raises ``ElementHandleError``.
        """
        await self._scrollIntoViewIfNeeded()
        center = await self._clickablePoint()
        x = center.get('x', 0)
        y = center.get('y', 0)
        await self._page.touchscreen.tap(x, y)

    async def focus(self) -> None:
        """Focus on this element."""
        await self.executionContext.evaluate('element => element.focus()', self)

    async def type(self, text: str, delay: float = 0) -> None:
        """Focus the element and then type text.

        Details see :meth:`pyppeteer.input.Keyboard.type` method.
        """
        await self.focus()
        await self._page.keyboard.type(text, delay)

    async def press(self, key: str, text: str = None, delay: float = 0) -> None:
        """Press ``key`` onto the element.

        This method focuses the element, and then uses
        :meth:`pyppeteer.input.keyboard.down` and
        :meth:`pyppeteer.input.keyboard.up`.

        :arg str key: Name of key to press, such as ``ArrowLeft``.

        This method accepts the following options:

        * ``text`` (str): If specified, generates an input event with this
          text.
        * ``delay`` (int|float): Time to wait between ``keydown`` and
          ``keyup``. Defaults to 0.
        """
        await self.focus()
        await self._page.keyboard.press(key, text=text, delay=delay)

    async def boundingBox(self) -> Optional[Dict[str, float]]:
        """Return bounding box of this element.

        If the element is not visible, return ``None``.

        This method returns dictionary of bounding box, which contains:

        * ``x`` (int): The X coordinate of the element in pixels.
        * ``y`` (int): The Y coordinate of the element in pixels.
        * ``width`` (int): The width of the element in pixels.
        * ``height`` (int): The height of the element in pixels.
        """
        result = await self._getBoxModel()

        if not result:
            return None

        quad = result['model']['border']
        x = min(quad[0], quad[2], quad[4], quad[6])
        y = min(quad[1], quad[3], quad[5], quad[7])
        width = max(quad[0], quad[2], quad[4], quad[6]) - x
        height = max(quad[1], quad[3], quad[5], quad[7]) - y
        return {'x': x, 'y': y, 'width': width, 'height': height}

    async def boxModel(self) -> Optional[Dict]:
        """Return boxes of element.

        Return ``None`` if element is not visible. Boxes are represented as an
        list of points; each Point is a dictionary ``{x, y}``. Box points are
        sorted clock-wise.

        Returned value is a dictionary with the following fields:

        * ``content`` (List[Dict]): Content box.
        * ``padding`` (List[Dict]): Padding box.
        * ``border`` (List[Dict]): Border box.
        * ``margin`` (List[Dict]): Margin box.
        * ``width`` (int): Element's width.
        * ``height`` (int): Element's height.
        """
        result = await self._getBoxModel()

        if not result:
            return None

        model = result.get('model', {})
        return {
            'content': self._fromProtocolQuad(model.get('content')),
            'padding': self._fromProtocolQuad(model.get('padding')),
            'border': self._fromProtocolQuad(model.get('border')),
            'margin': self._fromProtocolQuad(model.get('margin')),
            'width': model.get('width'),
            'height': model.get('height'),
        }

    async def screenshot(
        self,
        path: Union[str, Path] = None,
        type_: str = 'png',  # png or jpeg
        quality: int = None,  # 0 to 100
        fullPage: bool = False,
        omitBackground: bool = False,
        encoding: str = 'binary',
    ) -> bytes:
        """Take a screenshot of this element.

        If the element is detached from DOM, this method raises an
        ``ElementHandleError``.

        Available options are same as :meth:`pyppeteer.page.Page.screenshot`.
        """

        needsViewportReset = False
        boundingBox = await self.boundingBox()
        if not boundingBox:
            raise ElementHandleError('Node is either not visible or not an HTMLElement')

        original_viewport = copy.deepcopy(self._page.viewport)

        if boundingBox['width'] > original_viewport['width'] or boundingBox['height'] > original_viewport['height']:
            newViewport = {
                'width': max(original_viewport['width'], math.ceil(boundingBox['width'])),
                'height': max(original_viewport['height'], math.ceil(boundingBox['height'])),
            }
            new_viewport = copy.deepcopy(original_viewport)
            new_viewport.update(newViewport)
            await self._page.setViewport(new_viewport)
            needsViewportReset = True

        await self._scrollIntoViewIfNeeded()
        boundingBox = await self.boundingBox()
        if not boundingBox:
            raise ElementHandleError('Node is either not visible or not an HTMLElement')

        _obj = await self._client.send('Page.getLayoutMetrics')
        pageX = _obj['layoutViewport']['pageX']
        pageY = _obj['layoutViewport']['pageY']

        clip = {}
        clip.update(boundingBox)
        clip['x'] = clip['x'] + pageX
        clip['y'] = clip['y'] + pageY

        imageData = await self._page.screenshot(
            path=path,
            type_=type_,
            quality=quality,
            fullPage=fullPage,
            clip=clip,
            omitBackground=omitBackground,
            encoding=encoding,
        )

        if needsViewportReset:
            await self._page.setViewport(original_viewport)

        return imageData

    async def querySelector(self, selector: str) -> Optional['ElementHandle']:
        """Return first element which matches ``selector`` under this element.

        If no element matches the ``selector``, returns ``None``.
        """
        handle = await self.evaluateHandle('(element, selector) => element.querySelector(selector)', selector,)
        element = handle.asElement()
        if element:
            return element
        await handle.dispose()
        return None

    async def querySelectorAll(self, selector: str) -> List['ElementHandle']:
        """Return all elements which match ``selector`` under this element.

        If no element matches the ``selector``, returns empty list (``[]``).
        """
        arrayHandle = await self.executionContext.evaluateHandle(
            '(element, selector) => element.querySelectorAll(selector)', self, selector,
        )
        properties = await arrayHandle.getProperties()
        await arrayHandle.dispose()
        result = []
        for prop in properties.values():
            elementHandle = prop.asElement()
            if elementHandle:
                result.append(elementHandle)
        return result  # type: ignore

    async def querySelectorEval(self, selector: str, pageFunction: str, *args: JSFunctionArg) -> Any:
        """Run ``Page.querySelectorEval`` within the element.

        This method runs ``document.querySelector`` within the element and
        passes it as the first argument to ``pageFunction``. If there is no
        element matching ``selector``, the method raises
        ``ElementHandleError``.

        If ``pageFunction`` returns a promise, then wait for the promise to
        resolve and return its value.

        ``ElementHandle.Jeval`` is a shortcut of this method.

        Example:

        .. code:: python

            tweetHandle = await page.querySelector('.tweet')
            assert (await tweetHandle.querySelectorEval('.like', 'node => node.innerText')) == 100
            assert (await tweetHandle.Jeval('.retweets', 'node => node.innerText')) == 10
        """
        elementHandle = await self.querySelector(selector)
        if not elementHandle:
            raise ElementHandleError(f'Error: failed to find element matching selector "{selector}"')
        result = await self.executionContext.evaluate(pageFunction, elementHandle, *args)
        await elementHandle.dispose()
        return result

    async def querySelectorAllEval(self, selector: str, pageFunction: str, *args: JSFunctionArg) -> Any:
        """Run ``Page.querySelectorAllEval`` within the element.

        This method runs ``Array.from(document.querySelectorAll)`` within the
        element and passes it as the first argument to ``pageFunction``. If
        there is no element matching ``selector``, the method raises
        ``ElementHandleError``.

        If ``pageFunction`` returns a promise, then wait for the promise to
        resolve and return its value.

        Example:

        .. code:: html

            <div class="feed">
                <div class="tweet">Hello!</div>
                <div class="tweet">Hi!</div>
            </div>

        .. code:: python

            feedHandle = await page.J('.feed')
            assert (await feedHandle.JJeval('.tweet', '(nodes => nodes.map(n => n.innerText))')) == ['Hello!', 'Hi!']
        """
        arrayHandle = await self.executionContext.evaluateHandle(
            '(element, selector) => Array.from(element.querySelectorAll(selector))', self, selector
        )
        result = await self.executionContext.evaluate(pageFunction, arrayHandle, *args)
        await arrayHandle.dispose()
        return result

    async def xpath(self, expression: str) -> List['ElementHandle']:
        """Evaluate the XPath expression relative to this elementHandle.

        If there are no such elements, return an empty list.

        :arg str expression: XPath string to be evaluated.
        """
        arrayHandle = await self.executionContext.evaluateHandle(
            '''(element, expression) => {
                const document = element.ownerDocument || element;
                const iterator = document.evaluate(expression, element, null,
                    XPathResult.ORDERED_NODE_ITERATOR_TYPE);
                const array = [];
                let item;
                while ((item = iterator.iterateNext()))
                    array.push(item);
                return array;

            }''',
            self,
            expression,
        )
        properties = await arrayHandle.getProperties()
        await arrayHandle.dispose()
        result = []
        for property in properties.values():
            elementHandle = property.asElement()
            if elementHandle:
                result.append(elementHandle)
        return result

    #: alias to :meth:`xpath`
    Jx = xpath

    async def isIntersectingViewport(self) -> bool:
        """Return ``True`` if the element is visible in the viewport."""
        return await self.executionContext.evaluate(
            '''async element => {
            const visibleRatio = await new Promise(resolve => {
                const observer = new IntersectionObserver(entries => {
                    resolve(entries[0].intersectionRatio);
                    observer.disconnect();
                });
                observer.observe(element);
            });
            return visibleRatio > 0;
        }''',
            self,
        )


def computeQuadArea(quad: List[Dict]) -> float:
    area = 0
    for i in range(len(quad)):
        p1 = quad[i]
        p2 = quad[(i + 1) % len(quad)]
        area += (p1['x'] * p2['y'] - p2['x'] * p1['y']) / 2
    return abs(area)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Browser fetcher module.
Chromium is being downloaded from:
https://storage.googleapis.com/chromium-browser-snapshots
see full download instructions:
https://www.chromium.org/getting-involved/download-chromium
for latest version see:
https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Linux_x64%2FLAST_CHANGE?alt=media
"""

import logging
import os
import shutil
import struct
import sys
from distutils.util import strtobool
from io import BytesIO
from pathlib import Path
from typing import Any, List, Optional, Sequence, Tuple, Union, cast
from urllib import request
from zipfile import ZipFile
import encodings.cp437

import certifi
import urllib3
from tqdm import tqdm

from biolib.pyppeteer.pyppeteer import __chromium_revision__, __pyppeteer_home__
from biolib.pyppeteer.pyppeteer.models import Platform, RevisionInfo

if sys.version_info < (3, 8):
    from typing_inspect import get_args
else:
    from typing import get_args

logger = logging.getLogger(__name__)

DEFAULT_DOWNLOAD_HOST = 'https://storage.googleapis.com'
DOWNLOAD_HOST = os.environ.get('PYPPETEER_DOWNLOAD_HOST', DEFAULT_DOWNLOAD_HOST)


def get_archive_name(platform: Platform, revision: str) -> str:
    if platform == 'linux':
        return 'chrome-linux'
    if platform == 'mac':
        return 'chrome-mac'
    if platform in ('win32', 'win64'):
        return 'chrome-win' if int(revision) > 591479 else 'chrome-win32'
    raise ValueError(f'Unsupported platform argument: {platform}')


def download_url(platform: Platform, host: str, revision: str) -> str:
    windows_archive = 'chrome-win' if int(revision) > 591479 else 'chrome-win32'

    base_url = f'{host}/chromium-browser-snapshots'
    download_urls = {
        'linux': f'{base_url}/Linux_x64/{revision}/chrome-linux.zip',
        'mac': f'{base_url}/Mac/{revision}/chrome-mac.zip',
        'win32': f'{base_url}/Win/{revision}/{windows_archive}.zip',
        'win64': f'{base_url}/Win_x64/{revision}/{windows_archive}.zip',
    }

    return download_urls[platform]


def parse_folder_path(folder_path: Path) -> Tuple[Optional[str], Optional[str]]:
    name = folder_path.name
    splits: Sequence[Any] = name.split('-')
    if len(splits) != 2 or splits[0] not in get_args(Platform):  # type: ignore
        splits = (None, None)
    return cast(Tuple[Optional[str], Optional[str]], tuple(splits))


def download_file(url: str, zip_path: BytesIO) -> None:
    CHUNK_SIZE = 4096
    file_req = request.urlopen(url)
    NO_PROGRESS_BAR = bool(strtobool(os.environ.get('PYPPETEER_NO_PROGRESS_BAR', 'false')))


    progress_bar = tqdm(
        total=int(file_req.getheader('Content-Length', 0)),
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        disable=NO_PROGRESS_BAR,
    )
    for chunk in iter(lambda: file_req.read(CHUNK_SIZE), b''):
        progress_bar.update(CHUNK_SIZE)
        zip_path.write(chunk)
    progress_bar.close()


def extractZip(zip_file: BytesIO, folder_path: Path) -> None:
    with ZipFile(zip_file) as zf:
        zf.extractall(folder_path)


class BrowserFetcher:
    def __init__(
        self, projectRoot: Union[Path, str] = None, platform: Platform = None, host: str = None,
    ):
        self.downloadsFolder = Path(projectRoot or __pyppeteer_home__) / 'local-chromium'
        self.downloadHost = host or DEFAULT_DOWNLOAD_HOST
        plat = platform or sys.platform  # type: ignore
        if plat == 'darwin':
            plat = 'mac'
        elif plat == 'win32':
            # no really good way to detect system bittedness
            # (other options depend on the python interpreter bittedness == sys.bittedness)
            plat = plat.replace('32', str(struct.calcsize('P') * 8))
        assert plat in get_args(Platform), f'Unsupported platform: {platform}'  # type: ignore
        logger.info(f'platform auto detected: {plat}')
        self._platform = cast(Platform, plat)

    @property
    def platform(self) -> Platform:
        return self._platform

    def canDownload(self, revision: str) -> bool:
        url = download_url(self.platform, self.downloadHost, revision)
        return request.urlopen(request.Request(url, method='HEAD')) == 200

    def download(self, revision: Optional[str] = None) -> RevisionInfo:
        revision = revision or os.environ.get('PYPPETEER_CHROMIUM_REVISION') or __chromium_revision__
        url = download_url(self.platform, self.downloadHost, revision)
        folder_path = self._get_folder_path(revision)
        if folder_path.exists():
            return self.revision_info(revision)
        os.makedirs(self.downloadsFolder, exist_ok=True)
        with BytesIO() as zip_file_obj:
            download_file(url, zip_file_obj)
            extractZip(zip_file_obj, folder_path)
        revision_info = self.revision_info(revision)
        if revision_info:
            os.chmod(revision_info['executablePath'], 0o755)
        return revision_info

    def local_revisions(self) -> List[Path]:
        if not self.downloadsFolder.exists():
            return []
        result = []
        for file in [x for x in self.downloadsFolder.iterdir() if x.is_dir()]:
            platform, revision = parse_folder_path(file)
            if platform != self._platform or not revision:
                continue
            result.append(Path(revision))
        return result

    def remove(self, revision: str) -> None:
        f_path = self._get_folder_path(revision)
        assert f_path.exists(), f'Failed to remove: revision {revision} doesn\'t exist on the disk'
        shutil.rmtree(f_path)

    def revision_info(self, revision: str) -> RevisionInfo:
        folder_path = self._get_folder_path(revision)

        archive_name = get_archive_name(self._platform, revision)
        if self._platform == 'mac':
            executable_path = folder_path / archive_name / 'Chromium.app' / 'Contents' / 'MacOS' / 'Chromium'
        elif self._platform == 'linux':
            executable_path = folder_path / archive_name / 'chrome'
        elif self._platform in ('win32', 'win64'):
            executable_path = folder_path / archive_name / 'chrome.exe'
        else:
            raise RuntimeError(f'Unsupported platform: {self._platform}')

        url = download_url(self._platform, self.downloadHost, revision)
        local = folder_path.exists()

        return {
            'revision': revision,
            'executablePath': executable_path,
            'folderPath': folder_path,
            'local': local,
            'url': url,
        }

    def _get_folder_path(self, revision: str) -> Path:
        return self.downloadsFolder.joinpath(f'{self._platform}-{revision}')

    def can_download(self, revision: str) -> bool:
        url = download_url(self._platform, self.downloadHost, revision)
        http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())

        try:
            res = http.request('HEAD', url)
        except urllib3.exceptions.HTTPError as error:
            logger.error(error)
            return False

        return res.status == 200
